/** Parse and print engineering notation */
const prefixMap = {
  'y': 1e-24,
  'z': 1e-21,
  'a': 1e-18,
  'f': 1e-15,
  'p': 1e-12,
  'n': 1e-9,
  'u': 1e-6,
  'm': 1e-3,
  'k': 1e3,
  'M': 1e6,
  'G': 1e9,
  'T': 1e12,
  'P': 1e15,
  'E': 1e18,
  'Z': 1e21,
  'Y': 1e24
}
const REGEX_ENG = /^([+-])?\s*(\d*\.?\d+|\d+\.?\d*)?([yzafpnumkMGTPTEZY]|e[+-]?\d+)?$/

export function printEng (number) {
  const base = Math.floor(Math.floor(Math.log10(Math.fabs(number)) / 3))
  return base
}

export function parseEng (string) {
  const match = string.match(REGEX_ENG)
  if (!match) {
    return null
  }
  let num = 0
  let sign = 1
  let exp = 0
  if (match[1] === '-') {
    sign = -1
  }
  num = parseFloat(match[2])
  if (match[3] && match[3].length > 1) {
    exp = parseInt(match[3].substring(1))
  } else if (match[3]) {
    exp = prefixMap[match[3]]
  }
  return sign * num * Math.pow(10, exp)
}

/** Redux helpers */
import { call, put, takeEvery } from 'redux-sagas/effects'

export function genStuff (name, actionsList = ['create', 'update', 'remove'], sync = true, api = null) {
  const actionTypes = {}
  const actions = {}

  let actionList
  if (!sync) {
    let tmpActionList = []
    actionList.forEach(action => {
      tmpActionList.push(`${action}_request`)
      tmpActionList.push(`${action}_success`)
      tmpActionList.push(`${action}_failure`)
    })
    actionList = tmpActionList
  }

  /** Populate the actionTypes and actions */
  actionList.forEach(action => {
    const actionType = `${action.toUpperCase()}_${name.toUpperCase()}`
    const actionName = `${action}${name[0].toUpperCase()}${name.substring(1)}`
    actionTypes[actionType] = `${name}/${actionType}`
    actions[actionName] = (id, data) => { return { type: `${name}/${actionType}`, id, data } }
  })

  const state = []

  function * worker (action) {
    try {
      const data = yield call(api, payload)
      yield put({ type: '??_SUCCESS', data })
    } catch (error) {
      yield put({ type: '??_FAILURE', error })
    }
  }

  function * watcher (action) {
    actionList.forEach(action => {
      yield takeEvery(`${action.toUpperCase()}_REQUEST`, worker)
    })
  }

  const reducer = (sync) ? (state = [], action) => {
    let nextState
    switch (action.type) {
      case actionTypes[`CREATE_${name.toUpperCase()}`]:
        nextState = [ ...state, action.data ]
        return nextState
      case actionTypes[`UPDATE_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), { ...state[action.id], ...action.data }, ...state.slice(action.id + 1) ]
        return nextState
      case actionTypes[`REMOVE_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), ...state.slice(action.id + 1) ]
        return nextState
      default:
        return state
    }
  } : (state = [], action) => {
    let nextState
    switch (action.type) {
      case actionTypes[`CREATE_REQUEST_${name.toUpperCase()}`]:
        worker(action)
        return nextState
      case actionTypes[`CREATE_SUCCESS_${name.toUpperCase()}`]:
        nextState = [ ...state, action.data ]
        return nextState
      case actionTypes[`UPDATE_REQUEST_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), { ...state[action.id], ...action.data }, ...state.slice(action.id + 1) ]
        return nextState
      case actionTypes[`UPDATE_SUCCESS_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), { ...state[action.id], ...action.data }, ...state.slice(action.id + 1) ]
        return nextState
      case actionTypes[`REMOVE_REQUEST_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), ...state.slice(action.id + 1) ]
        return nextState
      case actionTypes[`REMOVE_SUCCESS_${name.toUpperCase()}`]:
        nextState = [ ...state.slice(0, action.id), ...state.slice(action.id + 1) ]
        return nextState
      default:
        return state
    }
  }

  return {
    actionTypes,
    actions,
    state,
    reducer,
    worker: (sync) ? null : worker,
    watcher: (sync) ? null : watcher
  }
}
